/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.main;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Set;

import bean.oag.OAG;
import bean.oag.context.Context;
import bean.oag.context.ContextSelector;
import bean.oag.context.RepresentativeContextSelector;
import bean.pta.Heap;
import bean.pta.PointsToAnalysis;
import bean.pta.doop.DoopPointsToAnalysis;
import bean.util.ANSIColor;
import bean.util.Pair;

public class DoopMain {
	
	private static final char SEP = '\t';
	private static final char EOL = '\n';
	
	public static void main(String[] args) throws Exception {
		String db = args[0];
		File contextFile = new File(args[1]);
		
		File dbDir = new File(db);
		System.out.println("Reading the points-to set from Doop ...");
		PointsToAnalysis pta = new DoopPointsToAnalysis(dbDir);
		
		System.out.println("Constructing object allocation graph (OAG) ...");
		long start = System.currentTimeMillis();
		OAG oag = new OAG(pta);
		long end = System.currentTimeMillis();
		System.out.printf(ANSIColor.BOLD + 
				"OAG construction: " +
				ANSIColor.GREEN +
				String.format("%.2fs", (end - start) / 1000F) +
				ANSIColor.RESET +
				"\n");
		
		System.out.println("Computing contexts...");
		start = System.currentTimeMillis();
		// The depth indicates the depth of heap context.
		// The method context has 1 more level than heap context.
		// Here depth is 1 which corresponds to 2-object-sensitive analysis
		// with 1 heap context.
		ContextSelector cs = new RepresentativeContextSelector(oag, 1);
		end = System.currentTimeMillis();
		System.out.printf(ANSIColor.BOLD + 
				"Context computation: " +
				ANSIColor.GREEN +
				String.format("%.2fs", (end - start) / 1000F) +
				ANSIColor.RESET +
				"\n");
		
		writeContext(cs, oag, contextFile);
	}
	
	private static void writeContext(ContextSelector cs, OAG oag, File contextFile)
			throws FileNotFoundException {
		// BeanHeapAllocation[allocHctx, allocator, heap] = newHctx
		PrintWriter writer = new PrintWriter(contextFile);
		oag.allNodes().forEach(node -> {
			Heap allocator = node.getHeap();
			Set<Context> ctxs = cs.contextsOf(allocator);
			for (Context ctx : ctxs) {
				Heap allocHctx = ctx.get(0);
				Set<Pair<Context, Heap>> csheaps = cs.allocatedBy(ctx, allocator);
				if (csheaps != null) {
					csheaps.forEach(csheap -> {
						Heap newHctx = csheap.getFirst().get(0);
						Heap heap = csheap.getSecond();
						writer.write(allocHctx.toString());
						writer.write(SEP);
						writer.write(allocator.toString());
						writer.write(SEP);
						writer.write(heap.toString());
						writer.write(SEP);
						writer.write(newHctx.toString());
						writer.write(EOL);
					});
				}
			}
		});
		writer.close();
	}
	
}
