/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.oag;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;
import java.util.stream.Collectors;

import bean.pta.CallGraph;
import bean.pta.Heap;
import bean.pta.Method;
import bean.pta.PointsToAnalysis;
import bean.pta.PointsToSet;
import bean.pta.Variable;
import bean.util.Pair;

/**
 * Implementation of Object Allocation Graph (OAG).
 * 
 */
public class OAG {

	private Set<Node> nodes = new HashSet<>();
	private Set<Node> rootNodes;
	private Set<Node> tailNodes;
	/** map a heap to its corresponding node in the OAG */
	private Map<Heap, Node> heap2node = new HashMap<>();

	public OAG(PointsToAnalysis pta) {
		buildOAG(pta);
		computeRootNodes();
		computeTailNodes();
	}

	public Set<Node> allNodes() {
		return nodes;
	}
	
	public Set<Node> rootNodes() {
		return rootNodes;
	}
	
	public Set<Node> tailNodes() {
		return tailNodes;
	}
	
	/**
	 * The nodes are computed on-the-fly and thus inefficient.
	 * @return The nodes whose number of in-degree is larger than 1.
	 */
	public Set<Node> junctionNodes() {
		return nodes.stream()
				.filter(n -> n.getInDegree() > 1)
				.collect(Collectors.toSet());
	}
	
	public Set<Node> getPredsOf(Node n) {
		return n.getPreds();
	}

	public Set<Node> getSuccsOf(Node n) {
		return n.getSuccs();
	}

	public int getInDegreeOf(Node n) {
		return n.getInDegree();
	}

	public int getOutDegreeOf(Node n) {
		return n.getOutDegree();
	}
	
	/**
	 * 
	 * @param source
	 * @param dest
	 * @return whether there is a path from source to target in the OAG
	 */
	public boolean reaches(Node source, Node dest) {
		Set<Node> reachableNodes = source.getReachableNodes();
		if (reachableNodes == null) {
			reachableNodes = computeReachableNodes(source);
			source.setReachableNodes(reachableNodes);
		}
		return reachableNodes.contains(dest);
	}
	
	/**
	 * @param source
	 * @return the nodes in OAG which can be reached from source
	 */
	private Set<Node> computeReachableNodes(Node source) {
		Set<Node> reachableNodes = new HashSet<>();
		Stack<Node> stack = new Stack<>();
		stack.push(source);
		while (!stack.isEmpty()) {
			Node node = stack.pop();
			if (reachableNodes.add(node)) {
				node.getSuccs().forEach(stack::add);
			}
		}
		return reachableNodes;
	}

	private void buildOAG(PointsToAnalysis pta) {
		PointsToSet p2set = pta.getPointsToSet();
		CallGraph cg = pta.getCallGraph();
		Map<Method, Set<Heap>> allocatorHeapsOfMethods =
				computeAllocatorsOfMethods(p2set, cg);
		p2set.allHeaps().forEach(heap -> {
			Node tgt = addNode(heap);
			Method inMethod = heap.inMethod();
			// Some special heap objects, e.g., string constants and class constants,
			// do not be allocated in any method.
			if (inMethod != null && allocatorHeapsOfMethods.containsKey(inMethod)) {
				Set<Heap> allocatorHeaps = allocatorHeapsOfMethods
						.get(inMethod);
				allocatorHeaps.forEach(allocatorHeap -> {
					Node src = addNode(allocatorHeap);
					addEdge(src, tgt);
				});
			}
		});
		addInitialAllocation(pta);
	}
	
	private void addInitialAllocation(PointsToAnalysis pta) {
		pta.getInitialAllocation().forEach(pair -> {
			Heap allocator = pair.getFirst();
			Heap heap = pair.getSecond();
			addEdge(addNode(allocator), addNode(heap));
		});
	}

	private Map<Method, Set<Heap>> computeAllocatorsOfMethods(
			PointsToSet p2set, CallGraph cg) {
		Map<Method, Set<Heap>> result = new HashMap<>();
		Queue<Pair<Method, Heap>> workList = new LinkedList<>();
		initWorkList(workList, p2set, cg);
		while (!workList.isEmpty()) {
			Pair<Method, Heap> pair = workList.poll();
			Method m = pair.getFirst();
			Heap h = pair.getSecond();
			if (!result.containsKey(m)) {
				result.put(m, new HashSet<>());
			}
			result.get(m).add(h);
			cg.getCalleesOf(m).stream().filter(callee -> callee.isStatic())
					.forEach(callee -> {
						Set<Heap> cacheHeaps = result.get(callee);
						if (cacheHeaps == null || !cacheHeaps.contains(h)) {
							workList.add(new Pair<>(callee, h));
						}
					});
		}
		return result;
	}

	/**
	 * Initialize every instance method
	 * 
	 * @param workList
	 */
	private void initWorkList(Queue<Pair<Method, Heap>> workList,
			PointsToSet p2set, CallGraph cg) {
		cg.allMethods()
				.stream()
				.filter(m -> m.isInstance())
				.forEach(m -> {
							Variable thisVar = m.getThis();
							p2set.getPointsToSetOf(thisVar).forEach(
									heap -> workList.add(new Pair<>(m, heap)));
						});
	}

	/**
	 * Add a node which represents the given heap object to the OAG. If the
	 * corresponding node already exists, then return the existing node.
	 * 
	 * @param heap
	 *            the one represented by the added node in the graph
	 * @return the new added node
	 */
	private Node addNode(Heap heap) {
		Node node = heap2node.get(heap);
		if (node == null) {
			node = new Node(heap);
			heap2node.put(heap, node);
			nodes.add(node);
		}
		return node;
	}

	/**
	 * Add a directed object allocation edge to the OAG.
	 * 
	 * @param src
	 *            source of the edge
	 * @param tgt
	 *            target of the edge
	 */
	private void addEdge(Node src, Node tgt) {
		src.addSucc(tgt);
	}
	
	private void computeRootNodes() {
		rootNodes = nodes.stream()
				.filter(node -> node.getInDegree() == 0)
				.collect(Collectors.toSet());
	}
	
	private void computeTailNodes() {
		tailNodes = nodes.stream()
				.filter(node -> node.getOutDegree() == 0)
				.collect(Collectors.toSet());
	}

}
