/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.oag.context;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.stream.Collectors;

import bean.oag.OAG;
import bean.oag.Node;
import bean.pta.Heap;
import bean.util.Triple;

/**
 * Context selector of BEAN. This selector aims to avoid redundant
 * context elements while selecting contexts.
 *
 */
public class RepresentativeContextSelector extends ContextSelector {

	public RepresentativeContextSelector(OAG oag) {
		super(oag);
	}
	
	public RepresentativeContextSelector(OAG oag, int depth) {
		super(oag, depth);
	}
	
	@Override
	protected void selectContext(OAG oag) {
		contextMap = new HashMap<>();
		oag.tailNodes().forEach(node -> {
			mergeContextMap(selectContext(oag, node));
		});
		oag.allNodes().forEach(node -> {
			if (!contextMap.containsKey(node.getHeap())) {
				mergeContextMap(selectContext(oag, node));
			}
		});
	}
	
	private Map<Heap, Set<Context>> selectContext(OAG oag, Node dest) {
		Map<Heap, Set<Context>> tempContextMap = new HashMap<>();
		Queue<Triple<Node, Context, Boolean>> worklist = new LinkedList<>();
		initialWorkList(worklist, oag, dest);
		while (!worklist.isEmpty()) {
			Triple<Node, Context, Boolean> triple = worklist.poll();
			Node node = triple.getFirst();
			Context ctx = triple.getSecond();
			boolean split = triple.getThird();
			Heap heap = node.getHeap();
			if (!tempContextMap.containsKey(heap)) {
				tempContextMap.put(heap, new HashSet<>());
			}
			if (tempContextMap.get(heap).add(ctx)) {
				Set<Node> reachSuccs = selectReachNodes(oag.getSuccsOf(node),
						dest, oag);
				final boolean isFork = reachSuccs.size() > 1;
				reachSuccs.forEach(succ -> {
					final boolean isJoinSucc = oag.getInDegreeOf(succ) > 1;
					Context newCtx;
					boolean succSplit = split;
					if (split && isJoinSucc && !ctx.contains(heap)) {
						newCtx = Context.newContext(ctx, heap, depth);
						succSplit = false;
					} else {
						newCtx = ctx;
					}
					if (isFork) {
						succSplit = true;
					}
					Set<Context> ctxs = tempContextMap.get(succ);
					if (ctxs == null || !ctxs.contains(newCtx)) {
						addAllocation(ctx, heap, newCtx, succ.getHeap());
						worklist.add(new Triple<>(succ, newCtx, succSplit));
					}
				});
			}
		}
		return tempContextMap;
	}
	
	private void mergeContextMap(Map<Heap, Set<Context>> anoContextMap) {
		anoContextMap.forEach((heap, ctxs) -> {
			if (contextMap.containsKey(heap)) {
				contextMap.get(heap).addAll(ctxs);
			} else {
				contextMap.put(heap, ctxs);
			}
		});
	}
	
	/**
	 * Select the nodes from a node set which can reach the destination node
	 * @param nodes
	 * @param dest
	 * @param oag
	 * @return
	 */
	private Set<Node> selectReachNodes(Set<Node> nodes, Node dest, OAG oag) {
		return nodes.stream()
				.filter(node -> oag.reaches(node, dest))
				.collect(Collectors.toSet());
	}
	
	private void initialWorkList(Queue<Triple<Node, Context, Boolean>> worklist,
			OAG oag, Node node) {
		Set<Node> reachRoots = selectReachNodes(oag.rootNodes(), node, oag);
		boolean split = reachRoots.size() > 1;
		Context emptyCtx = Context.emptyContext();
		reachRoots.forEach(root ->
				worklist.add(new Triple<>(root,
						Context.newContext(emptyCtx, root.getHeap(), depth),
						split)));
	}

}
