/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.pta.doop;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import bean.pta.CallGraph;
import bean.pta.Method;

class DoopCallGraph implements CallGraph {

	// Query predicates.
	private static final String REGULARCALL= "Stats:Simple:InsensCallGraphEdge";
	private static final String REFCALL = "_(from,to)<-ReflectiveCallGraphEdge(_,from,_,to).";
	private static final String NATIVECALL = "_(from,to)<-NativeCallGraphEdge(_,from,_,to).";
	private static final String MAINMETHOD = "MainMethodDeclaration";
	private static final String REACHABLE = "Reachable";
	private static final String IMPLICITREACHABLE = "ImplicitReachable";
	
	private Set<Method> methods = new HashSet<>();
	private Set<Method> implictEntries = new HashSet<>();
	private Method mainMethod;
	private Map<Method, Set<Method>> meth2callers = new HashMap<>();
	private Map<Method, Set<Method>> meth2callees = new HashMap<>();
	
	DoopCallGraph(DataBase db, MethodManager methManager)	throws IOException {
		addReachable(db, methManager);
		List<List<String>> cg = db.query(REGULARCALL);
		cg.addAll(db.query(REFCALL));
		cg.addAll(db.query(NATIVECALL));
		cg.forEach(list -> {
			String callsite = list.get(0);
			String targetSig = list.get(1);
			Method caller = methManager.getMethodOf(callsite);
			Method callee = methManager.getMethod(targetSig);
			methods.add(caller);
			methods.add(callee);
			addCallGraphEdge(caller, callee);
		});
		setMainMethod(db, methManager);
		computeImplicitEntries(db, methManager);
	}

	@Override
	public Set<Method> allMethods() {
		return methods;
	}

	@Override
	public Set<Method> implictEntryMethods() {
		return implictEntries;
	}

	@Override
	public Method mainMethod() {
		return mainMethod;
	}
	
	@Override
	public Set<Method> getCallersOf(Method m) {
		Set<Method> callers = this.meth2callers.get(m);
		return callers == null ? Collections.emptySet() : callers;
	}

	@Override
	public Set<Method> getCalleesOf(Method m) {
		Set<Method> callees = this.meth2callees.get(m);
		return callees == null ? Collections.emptySet() : callees;
	}
	
	private void addCallGraphEdge(Method caller, Method callee) {
		if (!meth2callers.containsKey(callee)) {
			meth2callers.put(callee, new HashSet<>());
		}
		meth2callers.get(callee).add(caller);
		if (!meth2callees.containsKey(caller)) {
			meth2callees.put(caller, new HashSet<>());
		}
		meth2callees.get(caller).add(callee);
	}
	
	private void addReachable(DataBase db, MethodManager methManager) throws IOException {
		db.query(REACHABLE)
				.stream()
				.map(list -> list.get(0))
				.map(methManager::getMethod)
				.forEach(methods::add);
		
	}
	
	private void setMainMethod(DataBase db, MethodManager methManager) throws IOException {
		String sig = db.query(MAINMETHOD).get(0).get(0);
		mainMethod = methManager.getMethod(sig);
	}
	
	private void computeImplicitEntries(DataBase db, MethodManager methManager)
			throws IOException {
		implictEntries = db.query(IMPLICITREACHABLE)
				.stream()
				.map(list -> list.get(0))
				.map(sig -> methManager.getMethod(sig))
				.collect(Collectors.toSet());
				
	}

}
