/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.pta.doop;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.Stack;
import java.util.stream.Collectors;

import bean.pta.CallGraph;
import bean.pta.Heap;
import bean.pta.Method;
import bean.pta.PointsToAnalysis;
import bean.pta.PointsToSet;
import bean.util.Pair;

public class DoopPointsToAnalysis implements PointsToAnalysis {

	private PointsToSet p2set;
	private CallGraph cg;
	private Set<Heap> specialHeaps;
	private Set<Pair<Heap, Heap>> initialAlloc;
	
	public DoopPointsToAnalysis(File dbDir) throws IOException {
		DataBase db = new DataBase(dbDir);
		VariableManager varManager = new VariableManager();
		MethodManager methManager = new MethodManager(db, varManager);
		HeapManager heapManager = new HeapManager(db, methManager);
		p2set = new DoopPointsToSet(db, varManager, heapManager);
		cg = new DoopCallGraph(db, methManager);
		initialSpecialHeaps(heapManager);
		computeInitialAllocation(heapManager);
	}
	
	@Override
	public PointsToSet getPointsToSet() {
		return p2set;
	}

	@Override
	public CallGraph getCallGraph() {
		return cg;
	}
	
	@Override
	public Set<Pair<Heap, Heap>> getInitialAllocation() {
		return initialAlloc;
	}
	
	private void initialSpecialHeaps(HeapManager heapManager) {
		specialHeaps = Arrays.asList(
				"<<class-initialization>>",
				"<<initial-context>>",
				"<<jvm-startup>>",
				"<<immutable>>",
				"<<main-thread>>",
				"<<main-thread-group>>",
				"<<system-thread-group>>"
				)
				.stream()
				.map(heapManager::getHeap)
				.collect(Collectors.toSet());
	}
	
	private void computeInitialAllocation(HeapManager heapManager) {
		initialAlloc = new HashSet<>();
		
		Heap immutableHeap = heapManager.getHeap("<<immutable>>");
		Set<Heap> immutableHeaps = computeImmutableHeaps();
		immutableHeaps
				.forEach(heap -> initialAlloc.add(new Pair<>(immutableHeap, heap)));
		
		Heap clinitHeap = heapManager.getHeap("<<class-initialization>>");
		computeClassInitializationHeaps()
				.stream()
				.filter(heap -> !immutableHeaps.contains(heap))
				.forEach(heap -> initialAlloc.add(new Pair<>(clinitHeap, heap)));
		
		Heap initHeap = heapManager.getHeap("<<initial-context>>");
		computeInitContextHeaps()
				.stream()
				.filter(heap -> !immutableHeaps.contains(heap))
				.forEach(heap -> initialAlloc.add(new Pair<>(initHeap, heap)));
		
		Heap jvmstartup = heapManager.getHeap("<<jvm-startup>>");
		computeImplicitReachableHeaps()
				.stream()
				.filter(heap -> !immutableHeaps.contains(heap))
				.forEach(heap -> initialAlloc.add(new Pair<>(jvmstartup, heap)));
		
		initialAlloc.add(new Pair<>(jvmstartup, heapManager.getHeap("<<main-thread>>")));
		initialAlloc.add(new Pair<>(jvmstartup, heapManager.getHeap("<<main-thread-group>>")));
		initialAlloc.add(new Pair<>(jvmstartup, heapManager.getHeap("<<system-thread-group>>")));
	}
	
	private Set<Heap> computeImmutableHeaps() {
		return p2set.allHeaps()
				.stream()
				.filter(heap -> !specialHeaps.contains(heap))
				.filter(heap -> heap.inMethod() == null)
				.collect(Collectors.toSet());
	}
	
	private Set<Heap> computeClassInitializationHeaps() {
		Set<Method> clinits = cg.allMethods()
				.stream()
				.filter(m -> m.getName().equals("<clinit>"))
				.collect(Collectors.toSet());
		return computeReachableHeaps(clinits);
	}
	
	private Set<Heap> computeInitContextHeaps() {
		return computeReachableHeaps(Collections.singleton(cg.mainMethod()));
	}
	
	private Set<Heap> computeImplicitReachableHeaps() {
		return computeReachableHeaps(cg.implictEntryMethods());
	}
	
	private Set<Heap> computeReachableHeaps(Set<Method> methods) {
		Set<Method> reachMethods = methods
				.stream()
				.map(this::computeReachableStaticMethods)
				.reduce(new HashSet<>(),
						(s1, s2) -> {
							s1.addAll(s2);
							return s1;
						});
		return p2set.allHeaps()
				.stream()
				.filter(h -> reachMethods.contains(h.inMethod()))
				.collect(Collectors.toSet());
	}
	
	private Set<Method> computeReachableStaticMethods(Method method) {
		Set<Method> reachableMethods = new HashSet<>();
		Stack<Method> stack = new Stack<>();
		stack.push(method);
		while (!stack.isEmpty()) {
			Method m = stack.pop();
			if (reachableMethods.add(m)) {
				cg.getCalleesOf(m).stream()
						.filter(meth -> meth.isStatic())
						.forEach(stack::add);
			}
		}
		return reachableMethods;
	}
	
}
