/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.pta.doop;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import bean.pta.Heap;
import bean.pta.Method;

class HeapManager {
	
	// Query predicates.
	private static final String SPECIAL_HEAP = "HeapAllocation:Special";
	private static final String MERGE_HEAP = "HeapAllocation:Merge";
	private static final String STRING_CONSTANTS = "StringConstant";
	private static final String REF_STRING_CONSTANTS = "ReflectionStringConstant";
	
	private static final String HEAP_IN = "_(heap,inmethod)<-"
			+ "AssignHeapAllocation(heap,_,inmethod).";
	private static final String REF_HEAP = "_(heap,callsite)<-"
			+ "ReflectiveHeapAllocation[callsite,_]=heap.";
	
	private static final String STRING_CONSTANT = "<<string-constant>>";
	
	private static boolean MERGE_NON_REFLECTION_STRINGS = true;
	
	private Map<String, String> heap2sig = new HashMap<>();
	private Map<String, Heap> heapName2heap = new HashMap<>();
	private Map<String, String> heap2callsite = new HashMap<>();
	
	private Map<String, String> heap2mergedheap = new HashMap<>();
	private Set<String> heapsToBeMerged = new HashSet<>();
	
	private MethodManager methManager;
	
	HeapManager(DataBase db, MethodManager methManager)
			throws IOException {
		Set<String> specialHeaps = db.query(SPECIAL_HEAP)
				.stream()
				.map(list -> list.get(0))
				.collect(Collectors.toSet());
		
		db.query(MERGE_HEAP)
				.forEach(list -> {
					String heap = list.get(0);
					String mergedHeap = list.get(1);
					heap2mergedheap.put(heap, mergedHeap);
					specialHeaps.add(mergedHeap);
					if (!heap.equals(mergedHeap)) {
						heapsToBeMerged.add(heap);
					}
				});
		
		if (MERGE_NON_REFLECTION_STRINGS) {
			mergeReflectionIrrelevantStrings(db, specialHeaps);
		}
		
		List<List<String>> heapIn = db.query(HEAP_IN);
		heapIn.forEach(list -> {
			String heapName = list.get(0);
			String methSig = list.get(1);
			// Filter out the containing methods of insensitive heaps and class constants
			// as their contexts are handled separately.
			if (!specialHeaps.contains(heapName) && !isClassConstant(heapName)) {
				heap2sig.put(heapName, methSig);
			}
		});
		
		List<List<String>> refHeap = db.query(REF_HEAP);
		refHeap.forEach(list -> heap2callsite.put(list.get(0), list.get(1)));
		
		this.methManager = methManager;
	}

	Heap getHeap(String heapName) {
		if (heapsToBeMerged.contains(heapName)) {
			return getHeap(heap2mergedheap.get(heapName));
		}
		Heap heap = heapName2heap.get(heapName);
		if (heap == null) {
			if (heap2sig.containsKey(heapName)) {
				String methSig = heap2sig.get(heapName);
				Method inMethod = methManager.getMethod(methSig);
				heap = new DoopHeap(heapName, inMethod);
			} else if (heap2callsite.containsKey(heapName)) {
				// In this case, the heap is created by a newInstance() call,
				// so the allocation site is the call site to newInstance().
				String callsite = heap2callsite.get(heapName);
				Method inMethod = methManager.getMethodOf(callsite);
				heap = new DoopHeap(heapName, inMethod);
			} else {
				heap = new DoopHeap(heapName, null);
			}
			heapName2heap.put(heapName, heap);
		}
		return heap;
	}
	
	/**
	 * Merge reflection-irrelevant string constants
	 * @param db
	 * @param specialHeaps 
	 * @throws IOException 
	 */
	private void mergeReflectionIrrelevantStrings(DataBase db, Set<String> specialHeaps)
			throws IOException {
		Set<String> refstrings = db.query(REF_STRING_CONSTANTS)
				.stream()
				.map(list -> list.get(0))
				.collect(Collectors.toSet());
		
		db.query(STRING_CONSTANTS)
				.stream()
				.map(list -> list.get(0))
				.filter(str -> !refstrings.contains(str))
				.forEach(str -> {
					heap2mergedheap.put(str, STRING_CONSTANT);
					specialHeaps.add(str);
					heapsToBeMerged.add(str);
				});
	}
	
	private boolean isClassConstant(String heapName) {
		return heapName.startsWith("<class ");
	}
	
}
