/* Bean - Making k-Object-Sensitive Pointer Analysis More Precise with Still k-Limiting
 *
 * Copyright (C) 2016 Tian Tan, Yue Li, Jingling Xue
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package bean.pta.doop;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import bean.pta.Method;

class MethodManager {

	// Query predicates.
	private static final String THISVAR = "ThisVar";
	private static final String CALLSITEIN = "_(callsite,inmethod)<-"
			+ "SpecialMethodInvocation:In(callsite,inmethod);"
			+ "VirtualMethodInvocation:In(callsite,inmethod);"
			+ "StaticMethodInvocation:In(callsite,inmethod).";
	private static final String OBJFINAL = "ObjectSupportsFinalize";
	
	private static final String THREADSTART0 = "<java.lang.Thread: void start0()>";
	
	private Map<String, String> sig2this = new HashMap<>();
	private Map<String, Method> sig2meth = new HashMap<>();
	private Map<String, String> callsite2sig = new HashMap<>();
	private Map<String, Method> callsite2meth = new HashMap<>();
	private Map<String, String> objfinal2sig = new HashMap<>();
	private VariableManager varManager;
	
	MethodManager(DataBase db, VariableManager varManager) throws IOException {
		List<List<String>> thisVar = db.query(THISVAR);
		thisVar.forEach(list -> sig2this.put(list.get(0), list.get(1)));
		List<List<String>> callsiteIn = db.query(CALLSITEIN);
		callsiteIn.forEach(list -> callsite2sig.put(list.get(0), list.get(1)));
		List<List<String>> objfinal = db.query(OBJFINAL);
		objfinal.forEach(list ->  objfinal2sig.put(list.get(0), list.get(1)));
		this.varManager = varManager;
		callsite2meth.put(THREADSTART0, getMethod(THREADSTART0));
	}
	
	Method getMethod(String sig) {
		Method method = sig2meth.get(sig);
		if (method == null) {
			String varName = sig2this.get(sig);
			if (varName == null) { // sig represents a static method
				method = new DoopMethod(sig, null, true);
			} else { // sig represents an instance method
				method = new DoopMethod(sig, varManager.getVariable(varName), false);
			}
			sig2meth.put(sig, method);
		}
		return method;
	}
	
	Method getMethodOf(String callsite) {
		if (!callsite2meth.containsKey(callsite)) {
			String sig = callsite2sig.get(callsite);
			if (sig == null) {
				// sig == null means that the given callsite is not a real callsite,
				// and it should be an allocation site which represents to an object
				// that supports finalization. Doop does this for modeling implicit
				// calls to java.lang.ref.Finalizer.register().
				// Refer to /logic/library/finalization.logic in Doop for details.
				sig = objfinal2sig.get(callsite);
			}
			callsite2meth.put(callsite, getMethod(sig));
		}
		return callsite2meth.get(callsite);
	}
	
}
